from dataclasses import dataclass, replace
import numpy as np
from .trigonometry import to_polar


@dataclass(frozen=True, slots=True)
class LampOrientation:
    x: float = 0.0
    y: float = 0.0
    z: float = 0.0
    angle: float = 0.0
    aimx: float = 0.0
    aimy: float = 0.0
    aimz: float = -1.0

    def __post_init__(self):
        if (self.aimx, self.aimy, self.aimz) == (self.x, self.y, self.z):
            raise ValueError("Aim point cannot equal position")

    def with_(self, **changes):
        return replace(self, **changes)

    @property
    def position(self):
        return np.array([self.x, self.y, self.z])

    @property
    def aim_point(self):
        return np.array([self.aimx, self.aimy, self.aimz])

    @property
    def heading(self):
        # xr, yr = self.aimx - self.x, self.aimy - self.y
        # return np.degrees(np.arctan2(yr, xr)) % 360
        dx, dy = self.aimx - self.x, self.aimy - self.y
        if np.isclose([dx, dy], 0.0).all():
            return 0.0
        return np.degrees(np.arctan2(dy, dx)) % 360

    @property
    def bank(self):
        dx, dy, dz = self.aimx - self.x, self.aimy - self.y, self.aimz - self.z
        norm = np.linalg.norm((dx, dy, dz))
        if norm == 0:
            return 0.0
        # cos(tilt) = -dz / |v| so tilt=0 → down, 180→up
        return np.degrees(np.arccos(np.clip(-dz / norm, -1.0, 1.0)))

    @property
    def rotation_matrix(self):
        # yaw1 = –heading, pitch = bank, yaw2 = –angle
        R1 = self._R_z(-self.heading)
        R2 = self._R_y(self.bank)
        R3 = self._R_z(-self.angle)
        return R3 @ R2 @ R1

    @property
    def inverse_rotation_matrix(self):
        return self.rotation_matrix.T

    def move(self, x=None, y=None, z=None):
        """Designate lamp position in cartesian space"""
        # determine new position   selected_lamp.
        x = self.x if x is None else x
        y = self.y if y is None else y
        z = self.z if z is None else z
        position = np.array([x, y, z])
        # update aim point based on new position
        diff = position - self.position
        aimx, aimy, aimz = self.aim_point + diff

        return self.with_(x=x, y=y, z=z, aimx=aimx, aimy=aimy, aimz=aimz)

    def rotate(self, angle):
        return self.with_(angle=angle)

    def aim(self, x=None, y=None, z=None):
        """aim lamp at a point in cartesian space"""
        aimx = self.aimx if x is None else x
        aimy = self.aimy if y is None else y
        aimz = self.aimz if z is None else z
        return self.with_(aimx=aimx, aimy=aimy, aimz=aimz)

    def tilt(self, tilt, dimensions=None, distance=None):
        """set the tilt, or bank"""
        return self.recalculate_aim_point(
            bank=tilt, dimensions=dimensions, distance=distance
        )

    def orient(self, orientation, dimensions=None, distance=None):
        """set the orientation, or heading"""
        return self.recalculate_aim_point(
            heading=orientation, dimensions=dimensions, distance=distance
        )

    def recalculate_aim_point(
        self, heading=None, bank=None, dimensions=None, distance=None
    ):
        """recalculate the aim point based on the heading and bank"""
        heading = self.heading if heading is None else heading
        bank = self.bank if bank is None else bank

        heading_rad = np.radians(heading)
        # Correcting bank angle for the pi shift
        bank_rad = np.radians(bank)  # - 180)
        # bank_rad = np.radians(np.clip(bank, 0, 180))

        # Convert from spherical to Cartesian coordinates
        dx = np.sin(bank_rad) * np.cos(heading_rad)
        dy = np.sin(bank_rad) * np.sin(heading_rad)
        dz = -np.cos(bank_rad)
        if dimensions is not None:
            distances = []
            dimx, dimy, dimz = dimensions
            if dx != 0:
                distances.append((dimx - self.x) / dx if dx > 0 else self.x / -dx)
            if dy != 0:
                distances.append((dimy - self.y) / dy if dy > 0 else self.y / -dy)
            if dz != 0:
                distances.append((dimz - self.z) / dz if dz > 0 else self.z / -dz)
            distance = min([d for d in distances])
        else:
            distance = 1 if distance is None else distance

        aimx, aimy, aimz = self.position + np.array([dx, dy, dz]) * distance
        return self.with_(aimx=aimx, aimy=aimy, aimz=aimz)

    def transform_to_world(self, coords, scale=1, which="cartesian"):
        """
        transform coordinates from the lamp frame of reference to the world
        Scale parameter should generally only be used for photometric_coords
        """
        coords = self.inverse_rotation_matrix @ coords.T
        coords = (coords / scale).T + self.position
        if which == "polar":
            return to_polar(*coords.T)
        elif which == "cartesian":
            return coords.T
        raise ValueError(f"`which` must be polar or cartesian, not {which}")

    def transform_to_lamp(self, coords, which="cartesian"):
        """
        transform coordinates to align with the lamp's coordinates
        """
        # coords = coords - self.position
        coords = self.rotation_matrix @ coords.T
        if which == "polar":
            return to_polar(*coords)
        elif which == "cartesian":
            return coords
        raise ValueError(f"`which` must be polar or cartesian, not {which}")

    # --------------- Internals --------------------------

    @staticmethod
    def _R_z(angle_deg):
        a = np.radians(angle_deg)
        return np.array(
            [[np.cos(a), -np.sin(a), 0], [np.sin(a), np.cos(a), 0], [0, 0, 1]]
        )

    @staticmethod
    def _R_y(angle_deg):
        a = np.radians(angle_deg)
        return np.array(
            [[np.cos(a), 0, np.sin(a)], [0, 1, 0], [-np.sin(a), 0, np.cos(a)]]
        )



import warnings
import inspect
from .lamp import Lamp
from .calc_zone import CalcPlane, CalcVol
from .room_plotter import RoomPlotter
from .room_dims import RoomDimensions
from .disinfection_calculator import DisinfectionCalculator
from .reflectance import ReflectanceManager
from .scene import Scene
from .io import load_room, save_room, export_room_zip, generate_report

VALID_UNITS = ["meters", "feet"]


class Room:
    """
    Represents a room containing lamps and calculation zones.

    The room is defined by its dimensions and can contain multiple lamps and
    different calculation zones for evaluating lighting conditions and other
    characteristics.
    """

    def __init__(
        self,
        x=None,
        y=None,
        z=None,
        units=None,
        standard=None,
        enable_reflectance=None,
        reflectances=None,
        reflectance_x_spacings=None,
        reflectance_y_spacings=None,
        reflectance_max_num_passes=None,
        reflectance_threshold=None,
        air_changes=None,
        ozone_decay_constant=None,
        precision=1,
        colormap="plasma",
        on_collision="increment",  # error | increment | overwrite
        unit_mode="auto",  # strict | auto
    ):

        ### Dimensions
        if units is not None:
            if units.lower() not in VALID_UNITS:
                raise KeyError(f"Invalid unit {units}")
            default = (
                [6.0, 4.0, 2.7] if units.lower() == "meters" else [20.0, 13.0, 9.0]
            )
        else:
            default = [6.0, 4.0, 2.7]

        self.dim = RoomDimensions(
            x if x is not None else default[0],
            y if y is not None else default[1],
            z if z is not None else default[2],
            "meters" if units is None else units.lower(),
        )

        ### Misc flags
        self.standard = standard or "ANSI IES RP 27.1-22 (ACGIH Limits)"
        self.air_changes = air_changes or 1.0
        self.ozone_decay_constant = ozone_decay_constant or 2.7
        self.precision = precision

        ### Scene - lamps and zones
        self.scene = Scene(
            dim=self.dim,
            on_collision=on_collision,
            unit_mode=unit_mode,
            colormap=colormap,
        )
        self.lamps = self.scene.lamps
        self.calc_zones = self.scene.calc_zones

        ### Reflectance
        self.enable_reflectance = (
            True if enable_reflectance is None else enable_reflectance
        )
        self.ref_manager = ReflectanceManager(
            x=self.dim.x,
            y=self.dim.y,
            z=self.dim.z,
            reflectances=reflectances,
            x_spacings=reflectance_x_spacings,
            y_spacings=reflectance_y_spacings,
            max_num_passes=reflectance_max_num_passes,
            threshold=reflectance_threshold,
        )

        ### Plotting and data extraction
        self._plotter = RoomPlotter(self)
        self._disinfection = DisinfectionCalculator(self)

        self.calc_state = {}
        self.update_state = {}

    def to_dict(self):
        data = {}
        data["x"] = self.dim.x
        data["y"] = self.dim.y
        data["z"] = self.dim.z
        data["units"] = self.units
        data["enable_reflectance"] = self.enable_reflectance
        data["reflectances"] = self.ref_manager.reflectances
        data["reflectance_x_spacings"] = self.ref_manager.x_spacings
        data["reflectance_y_spacings"] = self.ref_manager.y_spacings
        data["reflectance_max_num_passes"] = self.ref_manager.max_num_passes
        data["reflectance_threshold"] = self.ref_manager.threshold
        data["standard"] = self.standard
        data["air_changes"] = self.air_changes
        data["ozone_decay_constant"] = self.ozone_decay_constant
        data["precision"] = self.precision
        data["on_collision"] = self.scene.on_collision
        data["unit_mode"] = self.scene.unit_mode
        data["colormap"] = self.scene.colormap

        dct = self.__dict__.copy()
        data["lamps"] = {k: v.to_dict() for k, v in dct["lamps"].items()}
        data["calc_zones"] = {k: v.to_dict() for k, v in dct["calc_zones"].items()}
        return data

    @classmethod
    def from_dict(cls, data: dict):
        """recreate a room from a dict"""

        room_kwargs = list(inspect.signature(cls.__init__).parameters.keys())[1:]
        room = cls(**{k: v for k, v in data.items() if k in room_kwargs})

        for lampid, lamp in data["lamps"].items():
            room.add_lamp(Lamp.from_dict(lamp))

        for zoneid, zone in data["calc_zones"].items():
            if zone["calctype"] == "Plane":
                room.add_calc_zone(CalcPlane.from_dict(zone))
            elif zone["calctype"] == "Volume":
                room.add_calc_zone(CalcVol.from_dict(zone))
        return room

    def save(self, fname=None):
        """save all relevant parameters to a json file"""
        return save_room(self, fname)

    @classmethod
    def load(cls, filedata):
        """load a room from a json object"""
        return load_room(filedata)

    def export_zip(
        self,
        fname=None,
        include_plots=False,
        include_lamp_files=False,
        include_lamp_plots=False,
    ):
        """
        write the project file and all results files to a zip file. Optionally include
        extra files like lamp ies files, spectra files, and plots.
        """
        return export_room_zip(
            self,
            fname=fname,
            include_plots=include_plots,
            include_lamp_files=include_lamp_files,
            include_lamp_plots=include_lamp_plots,
        )

    def generate_report(self, fname=None):
        """generate a csv report of all the rooms contents and zone statistics"""
        return generate_report(self, fname)

    def get_calc_state(self):
        """
        Save all the features in the room that, if changed, will require re-calculation
        """

        room_state = [  # temp..this should be put with the ref_manager eventually
            self.enable_reflectance,
            tuple(self.ref_manager.x_spacings),
            tuple(self.ref_manager.y_spacings),
            self.ref_manager.max_num_passes,
            self.ref_manager.threshold,
        ]

        lamp_state = {}
        for key, lamp in self.scene.lamps.items():
            if lamp.enabled:
                lamp_state[key] = lamp.get_calc_state()

        zone_state = {}
        for key, zone in self.scene.calc_zones.items():
            if zone.calctype != "Zone" and zone.enabled:
                zone_state[key] = zone.get_calc_state()

        calc_state = {}
        calc_state["room"] = room_state
        calc_state["lamps"] = lamp_state
        calc_state["calc_zones"] = zone_state

        return calc_state

    def get_update_state(self):
        """
        Save all the features in the room that should NOT trigger
        a recalculation, only an update
        """

        room_state = [
            tuple(self.ref_manager.reflectances),
            self.units,
        ]

        lamp_state = {}
        for key, lamp in self.scene.lamps.items():
            lamp_state[key] = lamp.get_update_state()

        zone_state = {}
        for key, zone in self.scene.calc_zones.items():
            if zone.calctype != "Zone":
                zone_state[key] = zone.get_update_state()

        update_state = {}
        update_state["room"] = room_state
        update_state["lamps"] = lamp_state
        update_state["calc_zones"] = zone_state

        return update_state

    # --------------------- Misc flags -----------------------

    def set_standard(self, standard):
        """update the photobiological safety standard the Room is subject to"""
        self.standard = standard
        self.scene.update_standard_zones(standard)
        return self

    # --------------------- Reflectance ----------------------

    def set_reflectance(self, R, wall_id=None):
        """
        set the reflectance (a float between 0 and 1) for the reflective walls
        If wall_id is none, the value is set for all walls.
        """
        self.ref_manager.set_reflectance(R=R, wall_id=wall_id)
        return self

    def set_reflectance_spacing(self, x_spacing=None, y_spacing=None, wall_id=None):
        """
        set the spacing of the calculation points for the reflective walls
        If wall_id is none, the value is set for all walls.
        """
        self.ref_manager.set_spacing(
            x_spacing=x_spacing, y_spacing=y_spacing, wall_id=wall_id
        )
        return self

    def set_max_num_passes(self, max_num_passes):
        """set the maximum number of passes for the interreflection module"""
        self.ref_manager.max_num_passes = max_num_passes
        return self

    def set_reflectance_threshold(self, reflectance_threshold):
        """
        set the threshold percentage (a float between 0 and 1) for the
        interreflection module
        """
        self.ref_manager.threshold = reflectance_threshold
        return self

    # -------------- Dimensions and Units -----------------------

    def set_units(self, units, unit_mode="auto"):
        """set room units"""
        if units.lower() not in ["meters", "feet"]:
            raise KeyError("Valid units are `meters` or `feet`")
        self.dim = self.dim.with_(units=units)

        self.scene.dim = self.dim
        self.scene.update_standard_zones(self.standard)
        self.scene.to_units(unit_mode=unit_mode)
        return self

    def set_dimensions(self, x=None, y=None, z=None):
        """set room dimensions"""
        self.dim = self.dim.with_(x=x, y=y, z=z)
        self.ref_manager.update_dimensions(self.dim.x, self.dim.y, self.dim.z)
        self.scene.dim = self.dim
        self.scene.update_standard_zones(self.standard)
        return self

    @property
    def units(self) -> str:
        return self.dim.units

    @property
    def x(self) -> float:
        return self.dim.x

    @property
    def y(self) -> float:
        return self.dim.y

    @property
    def z(self) -> float:
        return self.dim.z

    @units.setter
    def units(self, value: str):
        self.set_units(value)

    @x.setter
    def x(self, value: float):
        self.set_dimensions(x=value)

    @y.setter
    def y(self, value: float):
        self.set_dimensions(y=value)

    @z.setter
    def z(self, value: float):
        self.set_dimensions(z=value)

    @property
    def dimensions(self) -> tuple[float, float, float]:
        return (self.dim.x, self.dim.y, self.dim.z)

    @property
    def volume(self) -> float:
        return self.dim.volume()

    # -------------------- Scene: lamps and zones ---------------------

    def add(self, *args, on_collision=None, unit_mode=None):
        """
        Add objects to the Scene.
        - If an object is a Lamp, it is added as a lamp.
        - If an object is a CalcZone, CalcPlane, or CalcVol, it is added as a calculation zone.
        - If an object is iterable, it is recursively processed.
        - Otherwise, a warning is printed.
        """
        self.scene.add(*args, on_collision=on_collision, unit_mode=unit_mode)
        return self

    def add_lamp(self, lamp, on_collision=None, unit_mode=None):
        """
        Add a lamp to the room scene
        """
        self.scene.add_lamp(lamp, on_collision=on_collision, unit_mode=unit_mode)
        return self

    def place_lamp(self, lamp, on_collision=None, unit_mode=None):
        """
        Position a lamp as far from other lamps and the walls as possible
        """
        self.scene.place_lamp(lamp, on_collision=on_collision, unit_mode=unit_mode)
        return self

    def place_lamps(self, *args, on_collision=None, unit_mode=None):
        """
        Place multiple lamps in the room, as far away from each other and the walls as possible
        """
        self.scene.place_lamps(*args, on_collision=on_collision, unit_mode=unit_mode)
        return self

    def remove_lamp(self, lamp_id):
        """Remove a lamp from the room scene"""
        self.scene.remove_lamp(lamp_id)
        return self

    def add_calc_zone(self, calc_zone, on_collision=None):
        """
        Add a calculation zone to the room
        """
        self.scene.add_calc_zone(calc_zone, on_collision=on_collision)
        return self

    def add_standard_zones(self, on_collision=None):
        """
        Add the special calculation zones SkinLimits, EyeLimits, and
        WholeRoomFluence to the room scene.
        If not overridden by user, standard zones are overwritten, not incremented
        """
        policy = on_collision or "overwrite"
        self.scene.add_standard_zones(self.standard, on_collision=policy)
        return self

    def remove_calc_zone(self, zone_id):
        """
        Remove a calculation zone from the room
        """
        self.scene.remove_calc_zone(zone_id)
        return self

    def set_colormap(self, colormap):
        """
        Set the room colormap
        """
        self.scene.set_colormap(colormap)
        return self

    def check_positions(self):
        """
        Verify the positions of all objects in the scene and return any warning messages
        """
        msgs = self.scene.check_positions()
        return msgs

    # -------------------------- Calculation ---------------------------

    def calculate(self, hard=False):
        """
        Triggers the calculation of lighting values in each calculation zone
        based on the current lamps in the room.

        If no updates have been made since the last calculate call that would
        require a full recalculation, either only an update will be performed
        or no recalculation will occur.

        If `hard` is True, this behavior is overriden and the full
        recalculation will be performed
        """

        valid_lamps = self.scene.get_valid_lamps()

        new_calc_state = self.get_calc_state()
        new_update_state = self.get_update_state()

        LAMP_RECALC = self.calc_state.get("lamps") != new_calc_state.get("lamps")
        REF_RECALC = self.calc_state.get("room") != new_calc_state.get("room")
        REF_UPDATE = self.update_state.get("room") != new_update_state.get("room")

        # calculate incidence on the surfaces if the reflectances or lamps have changed
        if (
            LAMP_RECALC or REF_RECALC or REF_UPDATE or hard
        ) and self.enable_reflectance:
            self.ref_manager.calculate_incidence(valid_lamps, hard=hard)

        ref_manager = self.ref_manager if self.enable_reflectance else None
        for name, zone in self.calc_zones.items():
            if zone.enabled:
                zone.calculate_values(
                    lamps=valid_lamps, ref_manager=ref_manager, hard=hard
                )
        # update calc states.
        self.calc_state = new_calc_state
        self.update_state = new_update_state

        # possibly this should be per-calc zone? idk maybe it's fine
        for lamp_id in valid_lamps.keys():
            new_calc_state = self.lamps[lamp_id].get_calc_state()
            self.lamps[lamp_id].calc_state = new_calc_state

        if len(valid_lamps) == 0:
            msg = "No valid lamps are present in the room--either lamps have been disabled, or filedata has not been provided."
            if len(self.lamps) == 0:
                msg = "No lamps are present in the room."
            warnings.warn(msg, stacklevel=3)

        return self

    def calculate_by_id(self, zone_id, hard=False):
        """calculate just the calc zone selected"""
        valid_lamps = self.scene.get_valid_lamps()

        if len(valid_lamps) > 0:
            new_calc_state = self.get_calc_state()
            new_update_state = self.get_update_state()

            LAMP_RECALC = self.calc_state.get("lamps") != new_calc_state.get("lamps")
            REF_RECALC = self.calc_state.get("room") != new_calc_state.get("room")
            REF_UPDATE = self.update_state.get("room") != new_update_state.get("room")

            # calculate incidence on the surfaces if the reflectances or lamps have changed
            if (
                LAMP_RECALC or REF_RECALC or REF_UPDATE or hard
            ) and self.enable_reflectance:
                self.ref_manager.calculate_incidence(valid_lamps, hard=hard)
            ref_manager = self.ref_manager if self.enable_reflectance else None
            self.calc_zones[zone_id].calculate_values(
                lamps=valid_lamps, ref_manager=ref_manager, hard=hard
            )
            self.calc_state = self.get_calc_state()
            self.update_state = self.get_update_state()
        return self

    # ------------------- Data and Plotting ----------------------

    def get_disinfection_data(self, zone_id="WholeRoomFluence"):
        """return the fluence_dict, dataframe, and violin plot"""
        return self._disinfection.get_disinfection_data(zone_id=zone_id)

    def plotly(self, fig=None, select_id=None, title=""):
        """return a plotly figure of all the room's components"""
        return self._plotter.plotly(fig=fig, select_id=select_id, title=title)



import numpy as np
import warnings


def to_polar(x, y, z, units="degrees"):
    """
    convert arrays of [x,y,z] points to arrays of [phi,theta,r] in degrees
    """
    r = np.sqrt(x ** 2 + y ** 2 + z ** 2)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        theta = np.arccos(-z / r)

    theta = np.nan_to_num(theta, nan=0)
    phi = np.arctan2(x, y)
    phi[np.where(phi < 0)] = phi[np.where(phi < 0)] + 2 * np.pi

    if units == "degrees":
        theta = np.degrees(theta)
        phi = np.degrees(phi)

    return np.array((theta, phi, r))


def to_cartesian(theta, phi, r):
    """
    convert from degrees polar to cartesian coordinates
    """
    theta_rad = np.radians(theta)
    phi_rad = np.radians(phi)
    x = r * np.sin(theta_rad) * np.sin(phi_rad)
    y = r * np.sin(theta_rad) * np.cos(phi_rad)
    z = r * np.cos(theta_rad)

    return np.array((x, y, z))


def attitude(coords, roll, pitch, yaw):

    roll = np.radians(roll)
    pitch = np.radians(pitch)
    yaw = np.radians(yaw)

    R_roll = np.array(
        [
            [1, 0, 0],
            [0, np.cos(roll), -np.sin(roll)],
            [0, np.sin(roll), np.cos(roll)],
        ]
    )

    R_pitch = np.array(
        [
            [np.cos(pitch), 0, np.sin(pitch)],
            [0, 1, 0],
            [-np.sin(pitch), 0, np.cos(pitch)],
        ]
    )

    R_yaw = np.array(
        [[np.cos(yaw), -np.sin(yaw), 0], [np.sin(yaw), np.cos(yaw), 0], [0, 0, 1]]
    )

    # Apply rotations
    coordst = coords.copy()
    coordst = np.dot(R_roll, coordst)
    coordst = np.dot(R_pitch, coordst)
    coordst = np.dot(R_yaw, coordst)
    newx, newy, newz = coordst

    return newx, newy, newz



import plotly.graph_objs as go
import numpy as np
from scipy.spatial import Delaunay
from .trigonometry import to_polar
from .calc_zone import CalcPlane, CalcVol
from .units import convert_units


class RoomPlotter:
    def __init__(self, room):
        self.room = room

    def plotly(
        self,
        fig=None,
        select_id=None,
        title="",
    ):
        """plot all"""
        if fig is None:
            fig = go.Figure()

        # first delete any extraneous traces
        lamp_ids = list(self.room.lamps.keys())
        aim_ids = [lampid + "_aim" for lampid in lamp_ids]
        zone_ids = list(self.room.calc_zones.keys())
        for active_ids in [lamp_ids, aim_ids, zone_ids]:
            self._remove_traces_by_ids(fig, active_ids)

        # plot lamps
        for lamp_id, lamp in self.room.lamps.items():
            if lamp.filedata is not None:
                fig = self._plot_lamp(lamp=lamp, fig=fig, select_id=select_id)
        for zone_id, zone in self.room.calc_zones.items():
            if isinstance(zone, CalcPlane):
                if zone.show_values and zone.values.sum() > 0:
                    fig = self._plot_plane_values(zone=zone, fig=fig)
                else:
                    fig = self._plot_plane(zone=zone, fig=fig, select_id=select_id)
            elif isinstance(zone, CalcVol):
                fig = self._plot_vol(zone=zone, fig=fig, select_id=select_id)

        x, y, z = self.room.dim.x, self.room.dim.y, self.room.dim.z

        # set views
        fig.update_layout(
            title=title,
            scene=dict(
                xaxis=dict(range=[0, x]),
                yaxis=dict(range=[0, y]),
                zaxis=dict(range=[0, z]),
                aspectratio=dict(
                    x=x / z,
                    y=y / z,
                    z=1,
                ),
            ),
            height=750,
            autosize=False,
            margin=go.layout.Margin(l=0, r=0, b=0, t=0, pad=0),
            legend=dict(
                x=0,
                y=1,
                yanchor="top",
                xanchor="left",
            ),
        )
        fig.add_annotation(
            text=f"Units: {self.room.units}",
            xref="paper",
            yref="paper",
            x=0,
            y=0,
            xanchor="left",
            yanchor="bottom",
            showarrow=False,
            font=dict(size=12, color="gray"),
            bgcolor="rgba(255,255,255,0.5)",
            borderpad=4,
        )
        fig.update_scenes(camera_projection_type="orthographic")
        return fig

    def _set_color(self, select_id, label, enabled):
        if not enabled:
            color = "#d1d1d1"  # grey
        elif select_id is not None and select_id == label:
            color = "#cc61ff"  # purple
        else:
            color = "#5e8ff7"  # blue
        return color

    def _update_trace_by_id(self, fig, trace_id, **updates):
        # Iterate through all traces
        for trace in fig.data:
            # Check if trace customdata matches the trace_id
            if trace.customdata and trace.customdata[0] == trace_id:
                # Update trace properties based on additional keyword arguments
                for key, value in updates.items():
                    setattr(trace, key, value)

    def _remove_traces_by_ids(self, fig, active_ids):
        # Convert fig.data, which is a tuple, to a list to allow modifications
        traces = list(fig.data)

        # Iterate in reverse to avoid modifying the list while iterating
        for i in reversed(range(len(traces))):
            trace = traces[i]
            # Check if the trace's customdata is set and its ID is not in the list of active IDs
            if trace.customdata and trace.customdata[0] not in active_ids:
                del traces[i]  # Remove the trace from the list
        fig.data = traces

    def _plot_lamp(self, lamp, fig, select_id=None, color="#cc61ff"):
        """plot lamp as a photometric web"""

        init_scale = convert_units(self.room.units, "meters", lamp.values.max())
        coords = lamp.transform_to_world(lamp.photometric_coords, scale=init_scale)
        scale = lamp.get_total_power() / 120
        coords = (coords.T - lamp.position) * scale + lamp.surface.position
        x, y, z = coords.T

        Theta, Phi, R = to_polar(*lamp.photometric_coords.T)
        tri = Delaunay(np.column_stack((Theta.flatten(), Phi.flatten())))
        lampcolor = self._set_color(select_id, label=lamp.lamp_id, enabled=lamp.enabled)

        lamptrace = go.Mesh3d(
            x=x,
            y=y,
            z=z,
            i=tri.simplices[:, 0],
            j=tri.simplices[:, 1],
            k=tri.simplices[:, 2],
            color=lampcolor,
            opacity=0.4,
            name=lamp.name,
            customdata=["lamp_" + lamp.lamp_id],
            legendgroup="lamp_" + lamp.lamp_id,
            # legendgroup="lamps",
            # legendgrouptitle_text="Lamps",
            showlegend=True,
        )
        xi, yi, zi = lamp.surface.position
        xia, yia, zia = lamp.aim_point
        aimtrace = go.Scatter3d(
            x=[xi, xia],
            y=[yi, yia],
            z=[zi, zia],
            mode="lines",
            line=dict(color="black", width=2, dash="dash"),
            name=lamp.name,
            customdata=["lamp_" + lamp.lamp_id + "_aim"],
            legendgroup="lamp_" + lamp.lamp_id,
            # legendgroup="lamps",
            # legendgrouptitle_text="Lamps",
            showlegend=False,
        )
        xs, ys, zs = lamp.surface.surface_points.T
        surfacetrace = go.Scatter3d(
            x=xs,
            y=ys,
            z=zs,
            mode="markers",
            marker=dict(size=2, color=lampcolor),
            opacity=0.9,
            name=lamp.name,
            customdata=["lamp_" + lamp.lamp_id + "_surface"],
            legendgroup="lamp_" + lamp.lamp_id,
            # legendgroup="lamps",
            # legendgrouptitle_text="Lamps",
            showlegend=False,
        )

        traces = [trace.customdata[0] for trace in fig.data]
        if lamptrace.customdata[0] not in traces:
            fig.add_trace(lamptrace)
            fig.add_trace(aimtrace)
            fig.add_trace(surfacetrace)
        else:
            self._update_trace_by_id(
                fig,
                lamp.lamp_id,
                x=x,
                y=y,
                z=z,
                i=tri.simplices[:, 0],
                j=tri.simplices[:, 1],
                k=tri.simplices[:, 2],
                color=lampcolor,
                name=lamp.name,
            )

            self._update_trace_by_id(
                fig,
                lamp.lamp_id + "_aim",
                x=[xi, xia],
                y=[yi, yia],
                z=[zi, zia],
            )

            self._update_trace_by_id(fig, lamp.lamp_id + "_surface", x=xs, y=ys, z=zs)
        return fig

    def _plot_plane(self, zone, fig, select_id=None):
        """plot a calculation plane"""
        zonecolor = self._set_color(select_id, zone.zone_id, zone.enabled)
        x, y, z = zone.coords.T
        zonetrace = go.Scatter3d(
            x=x,
            y=y,
            z=z,
            mode="markers",
            marker=dict(size=2, color=zonecolor),
            opacity=0.5,
            # legendgroup="planes",
            # legendgrouptitle_text="Calculation Planes",
            showlegend=True,
            name=zone.name,
            customdata=["zone_" + zone.zone_id],
        )
        traces = [trace.name for trace in fig.data]
        if zonetrace.name not in traces:
            fig.add_trace(zonetrace)
        else:
            self._update_trace_by_id(
                fig,
                zone.zone_id,
                x=x,
                y=y,
                z=z,
                marker=dict(size=2, color=zonecolor),
            )

        return fig

    def _plot_plane_values(self, zone, fig):
        if zone.ref_surface == "xy":
            X, Y = np.meshgrid(zone.xp, zone.yp, indexing="ij")
            Z = np.full_like(X, zone.height)
        elif zone.ref_surface == "xz":
            X, Z = np.meshgrid(zone.xp, zone.yp, indexing="ij")
            Y = np.full_like(X, zone.height)
        elif zone.ref_surface == "yz":
            Y, Z = np.meshgrid(zone.xp, zone.yp, indexing="ij")
            X = np.full_like(X, zone.height)
        zone_value_trace = go.Surface(
            x=X,
            y=Y,
            z=Z,
            surfacecolor=zone.values,
            colorscale=self.room.scene.colormap,
            showscale=False,
            # colorbar=None,
            # legendgroup="planes",
            # legendgrouptitle_text="Calculation Planes",
            showlegend=True,
            name=zone.name,
            customdata=["zone_" + zone.zone_id],
        )
        traces = [trace.name for trace in fig.data]
        if zone_value_trace.name not in traces:
            fig.add_trace(zone_value_trace)
        else:
            self._update_trace_by_id(
                fig,
                zone.zone_id,
                x=X,
                y=Y,
                z=Z,
                surfacecolor=zone.values,
            )
        return fig

    def _plot_vol(self, zone, fig, select_id=None):
        # Define the vertices of the rectangular prism
        (x1, y1, z1), (x2, y2, z2) = zone.dimensions
        vertices = [
            (x1, y1, z1),  # 0
            (x2, y1, z1),  # 1
            (x2, y2, z1),  # 2
            (x1, y2, z1),  # 3
            (x1, y1, z2),  # 4
            (x2, y1, z2),  # 5
            (x2, y2, z2),  # 6
            (x1, y2, z2),  # 7
        ]

        # Define edges by vertex indices
        edges = [
            (0, 1),
            (1, 2),
            (2, 3),
            (3, 0),  # Bottom face
            (4, 5),
            (5, 6),
            (6, 7),
            (7, 4),  # Top face
            (0, 4),
            (1, 5),
            (2, 6),
            (3, 7),  # Side edges
        ]

        # Create lists for x, y, z coordinates
        x_coords = []
        y_coords = []
        z_coords = []

        # Append coordinates for each edge, separated by None to create breaks
        for v1, v2 in edges:
            x_coords.extend([vertices[v1][0], vertices[v2][0], None])
            y_coords.extend([vertices[v1][1], vertices[v2][1], None])
            z_coords.extend([vertices[v1][2], vertices[v2][2], None])

        zonecolor = self._set_color(select_id, label=zone.zone_id, enabled=zone.enabled)
        # Create a single trace for all edges
        zonetrace = go.Scatter3d(
            x=x_coords,
            y=y_coords,
            z=z_coords,
            mode="lines",
            line=dict(color=zonecolor, width=5, dash="dot"),
            # legendgroup="volumes",
            # legendgrouptitle_text="Calculation Volumes",
            name=zone.name,
            customdata=["zone_" + zone.zone_id],
        )
        traces = [trace.name for trace in fig.data]
        if zonetrace.name not in traces:
            fig.add_trace(zonetrace)
        else:
            self._update_trace_by_id(
                fig,
                zone.zone_id,
                x=x_coords,
                y=y_coords,
                z=z_coords,
                line=dict(color=zonecolor, width=5, dash="dot"),
            )
        # fluence isosurface
        if zone.values is not None and zone.show_values:
            X, Y, Z = np.meshgrid(*zone.points, indexing="ij")
            x, y, z = X.flatten(), Y.flatten(), Z.flatten()
            values = zone.values.flatten()
            isomin = zone.values.mean() / 2
            if zone.name + " Values" not in traces:
                zone_value_trace = go.Isosurface(
                    x=x,
                    y=y,
                    z=z,
                    value=values,
                    isomin=isomin,
                    surface_count=3,
                    opacity=0.25,
                    showscale=False,
                    colorbar=None,
                    colorscale=self.room.scene.colormap,
                    name=zone.name + " Values",
                    customdata=["zone_" + zone.zone_id + "_values"],
                    # legendgroup="volumes",
                    # legendgrouptitle_text="Calculation Volumes",
                    showlegend=True,
                )
                fig.add_trace(zone_value_trace)
            else:
                self._update_trace_by_id(
                    fig,
                    zone.zone_id,
                    x=x,
                    y=y,
                    z=z,
                    values=values,
                    isomin=isomin,
                )

        return fig



from collections.abc import Iterable
from collections import defaultdict
import warnings
from matplotlib import colormaps
from .room_dims import RoomDimensions
from .lamp import Lamp
from .calc_zone import CalcZone, CalcPlane, CalcVol
from .lamp_helpers import new_lamp_position


class Scene:
    def __init__(
        self, dim: RoomDimensions, unit_mode: str, on_collision: str, colormap: str
    ):
        self.dim = dim
        self.unit_mode: str = unit_mode  # "strict" → raise; "auto" → convert in place
        self.on_collision: str = on_collision  # error | increment | overwrite"
        self.colormap: str = colormap

        self.lamps: dict[str, Lamp] = {}
        self.calc_zones: dict[str, CalcZone] = {}

        # for generating unique IDs
        self._lamp_counter = defaultdict(int)
        self._zone_counter = defaultdict(int)

    def add(self, *args, on_collision=None, unit_mode=None):
        """
        Add objects to the Scene.
        - If an object is a Lamp, it is added as a lamp.
        - If an object is a CalcZone, CalcPlane, or CalcVol, it is added as a calculation zone.
        - If an object is iterable, it is recursively processed.
        - Otherwise, a warning is printed.
        """

        for obj in args:
            if isinstance(obj, Lamp):
                self.add_lamp(obj, on_collision=on_collision, unit_mode=unit_mode)
            elif isinstance(obj, (CalcZone, CalcPlane, CalcVol)):
                self.add_calc_zone(obj, on_collision=on_collision)
            elif isinstance(obj, dict):
                self.add(*obj.values(), on_collision=on_collision)
            elif isinstance(obj, Iterable) and not isinstance(obj, (str, bytes)):
                self.add(
                    *obj, on_collision=on_collision
                )  # Recursively process other iterables
            else:
                msg = f"Cannot add object of type {type(obj).__name__} to Room."
                warnings.warn(msg, stacklevel=3)

    def add_lamp(self, lamp, base_id="Lamp", on_collision=None, unit_mode=None):
        """Add a lamp to the room"""
        
        lamp_id = self._get_id(
            mapping=self.lamps,
            obj_id=lamp.lamp_id,
            base_id=base_id,
            counter=self._lamp_counter,
            on_collision=on_collision,
        )
        lamp.lamp_id = lamp_id
        if lamp.name is None:
            lamp.name = lamp_id
        self.lamps[lamp_id] = self._check_lamp(lamp, unit_mode=unit_mode)
        return self

    def place_lamp(self, lamp, on_collision=None, unit_mode=None):
        """
        Position a lamp as far from other lamps and the walls as possible
        """
        idx = len(self.lamps) + 1
        x, y = new_lamp_position(idx, self.dim.x, self.dim.y)
        lamp.move(x, y, self.dim.z)
        self.add_lamp(lamp, on_collision=on_collision, unit_mode=unit_mode)

    def place_lamps(self, *args, on_collision=None, unit_mode=None):
        """place multiple lamps in the room, as far away from each other and the walls as possible"""
        for obj in args:
            if isinstance(obj, Lamp):
                self.place_lamp(obj, on_collision=on_collision, unit_mode=unit_mode)
            else:
                msg = f"Cannot add object of type {type(obj).__name__} to Room."
                warnings.warn(msg, stacklevel=3)

    def remove_lamp(self, lamp_id):
        """Remove a lamp from the scene"""
        self.lamps.pop(lamp_id, None)

    def set_colormap(self, colormap: str):
        """Set the scene's colormap"""
        if colormap not in list(colormaps):
            warnings.warn(f"{colormap} is not a valid colormap.")
        else:
            self.colormap = colormap
            for zone in self.calc_zones.values():
                zone.colormap = self.colormap

    def add_calc_zone(self, zone, base_id=None, on_collision=None):
        """
        Add a calculation zone to the scene
        """
        if base_id is None:
            base_id = "Calc" + zone.calctype

        zone_id = self._get_id(
            mapping=self.calc_zones,
            obj_id=zone.zone_id,
            base_id=base_id,
            counter=self._zone_counter,
            on_collision=on_collision,
        )
        zone.zone_id = zone_id
        if zone.name is None:
            zone.name = zone_id
        zone.colormap = self.colormap
        self.calc_zones[zone_id] = self._check_zone(zone)

    def remove_calc_zone(self, zone_id):
        """remove calculation zone from scene"""
        self.calc_zones.pop(zone_id, None)

    def add_standard_zones(self, standard, *, on_collision=None):
        """
        Add the special calculation zones SkinLimits, EyeLimits, and
        WholeRoomFluence to the scene
        """
        standard_zones = [
            CalcVol(
                zone_id="WholeRoomFluence",
                name="Whole Room Fluence",
                show_values=False,
            ),
            CalcPlane(
                zone_id="EyeLimits",
                name="Eye Dose (8 Hours)",
                dose=True,
                hours=8,
            ),
            CalcPlane(
                zone_id="SkinLimits",
                name="Skin Dose (8 Hours)",
                dose=True,
                hours=8,
            ),
        ]

        self.add(standard_zones, on_collision=on_collision)
        # sets the height and field of view parameters
        self.update_standard_zones(standard)

    def update_standard_zones(self, standard):
        """
        update the standard safety calculation zones based on the current
        standard, units, and room dimensions
        """
        if "UL8802" in standard:
            height = 1.9 if self.dim.units == "meters" else 6.25
            skin_horiz = False
            eye_vert = False
            fov_vert = 180
        else:
            height = 1.8 if self.dim.units == "meters" else 5.9
            skin_horiz = True
            eye_vert = True
            fov_vert = 80

        if "SkinLimits" in self.calc_zones.keys():
            zone = self.calc_zones["SkinLimits"]
            zone.set_dimensions(x2=self.dim.x, y2=self.dim.y)
            zone.set_height(height)
            zone.horiz = skin_horiz
        if "EyeLimits" in self.calc_zones.keys():
            zone = self.calc_zones["EyeLimits"]
            zone.set_dimensions(x2=self.dim.x, y2=self.dim.y)
            zone.set_height(height)
            zone.fov_vert = fov_vert
            zone.vert = eye_vert
        if "WholeRoomFluence" in self.calc_zones.keys():
            zone = self.calc_zones["WholeRoomFluence"]
            zone.set_dimensions(x2=self.dim.x, y2=self.dim.y, z2=self.dim.z)

    def check_positions(self):
        """
        verify the positions of all objects in the scene and return any warning messages
        """
        msgs = []
        for lamp_id, lamp in self.lamps.items():
            msgs.append(self._check_lamp_position(lamp))
        for zone_id, zone in self.calc_zones.items():
            msgs.append(self._check_zone_position(zone))
        return msgs

    def get_valid_lamps(self):
        """return all the lamps that can participate in a calculation"""
        return {
            k: v for k, v in self.lamps.items() if v.enabled and v.filedata is not None
        }

    def to_units(self, unit_mode=None):
        """
        ensure that all lamps in the state have the correct units, or raise an error
        in strict mode
        """
        for lamp in self.lamps.values():
            self._check_lamp_units(lamp, unit_mode=unit_mode)

    # --------------------------- internals ----------------------------

    def _get_id(self, mapping, obj_id, base_id, counter, on_collision=None):
        """generate an ID for a lamp or calc zone object"""
        policy = on_collision or self.on_collision
        if obj_id is None:
            return self._unique_id(base_id, counter)
        elif obj_id in mapping:
            if policy=="error":
                raise ValueError(f"'{obj_id}' already exists")
            elif policy == "overwrite":
                return str(obj_id) # does not bump unique_id counter
        return self._unique_id(str(obj_id), counter) # increment counter

    def _unique_id(self, base: str, counter: defaultdict) -> str:
        counter[base] += 1
        return base if counter[base] == 1 else f"{base}-{counter[base]}"

    def _check_lamp(self, lamp, unit_mode=None):
        """check lamp position and units"""
        if not isinstance(lamp, Lamp):
            raise TypeError(f"Must be type Lamp, not {type(lamp)}")
        self._check_lamp_position(lamp)
        self._check_lamp_units(lamp, unit_mode)
        return lamp

    def _check_lamp_units(self, lamp, unit_mode=None):
        """convert lamp units, or raise error in strict mode"""
        policy = unit_mode or self.unit_mode
        if lamp.surface.units != self.dim.units:
            if policy == "strict":
                raise ValueError(
                    f"Lamp {lamp.lamp_id} is in {lamp.surface.units}, "
                    f"room is {self.dim.units}"
                )
            lamp.set_units(self.dim.units)

    def _check_zone(self, zone):
        if not isinstance(zone, (CalcZone, CalcPlane, CalcVol)):
            raise TypeError(f"Must be CalcZone, CalcPlane, or CalcVol not {type(zone)}")
        # self._check_zone_position(zone)
        return zone

    def _check_lamp_position(self, lamp):
        return self._check_position(lamp.position, lamp.name)

    def _check_zone_position(self, calc_zone):
        if isinstance(calc_zone, CalcPlane):
            dimensions = [calc_zone.x2, calc_zone.y2]
        elif isinstance(calc_zone, CalcVol):
            dimensions = [calc_zone.x2, calc_zone.y2, calc_zone.z2]
        elif isinstance(calc_zone, CalcZone):
            # this is a hack; a generic CalcZone is just a placeholder
            dimensions = self.dim.dimensions()
        return self._check_position(dimensions, calc_zone.name)

    def _check_position(self, dimensions, obj_name):
        """
        Method to check if an object's dimensions exceed the room's boundaries.
        """
        msg = None
        for coord, roomcoord in zip(dimensions, self.dim.dimensions()):
            if coord > roomcoord:
                msg = f"{obj_name} exceeds room boundaries!"
                warnings.warn(msg, stacklevel=2)
        return msg
